"""The scrippy_core.workspace sub-module provides the necessary objects for workspace management."""
import os
import shutil
import pathlib
from scrippy_core.context import GlobalContext
from scrippy_core.error_handler import ScrippyCoreError


class Workspace:
  """The Workspace class provides a workspace with a directory based on the name passed as an argument to the constructor.

  The workspace will be created in the SCRIPPY_TMPDIR directory."""

  def __init__(self, name):
    env = GlobalContext().get_env()
    self.logger = GlobalContext().get_logger()
    self.path = pathlib.Path(os.path.join(env.get("tmpdir"), name))

  def __enter__(self):
    """Entry point"""
    self.logger.info(f"[+] Using workspace: {self.path}")
    try:
      self.path.mkdir(mode=0o750, parents=True, exist_ok=False)
    except Exception as err:
      err_msg = f"Unable to create workspace: [{err.__class__.__name__}] {err}"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    return str(self.path)

  def __exit__(self, kind, value, traceback):
    """Exit point."""
    self.logger.info(f"[+] Workspace deletion: {self.path}")
    try:
      shutil.rmtree(self.path)
    except Exception as err:
      err_msg = f"Error while deleting workspace: [{err.__class__.__name__}] {err}"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def __eq__(self, other):
    """Override default implementation."""
    if isinstance(other, Workspace):
      return self.path == other.path
    return False

  def __str__(self):
    return str(self.path)



"""The scrippy_core.arguments sub-module provides all the necessary objects and functions for handling arguments passed to the scripts."""
import argparse
import argcomplete
from scrippy_core.scriptinfo import ScriptInfo
from scrippy_core.context import GlobalContext
from scrippy_core.arguments.logaction import LogAction
from scrippy_core.error_handler import ScrippyCoreError
from scrippy_core.arguments.sourceaction import SourceAction
from scrippy_core.arguments.historyaction import HistoryAction


class Parser:
  """The Parser class enables the analysis of arguments passed to the scripts."""

  def __init__(self):
    """Parser instantiation."""
    self.logger = GlobalContext().get_logger()
    self.required = {'true': True, 'false': False}
    self.parser = argparse.ArgumentParser(prog=ScriptInfo.get_filename(),
                                          formatter_class=argparse.RawTextHelpFormatter)
    self._add_default_args()
    self._add_args()
    argcomplete.autocomplete(self.parser)
    self.args = self.parser.parse_args()
    self._check_implies(self.args)
    self._check_conflicts(self.args)

  def _get_description(self):
    try:
      description = [line.split(':')[1].strip() for line in ScriptInfo.get_doc().splitlines() if line.strip().startswith("@description")]
      return description[0]
    except IndexError as err:
      err_msg = "Missing script description"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def _get_author(self):
    try:
      author = [line.split(':')[1].strip() for line in ScriptInfo.get_doc().splitlines() if line.strip().startswith("@author")]
      return author[0]
    except IndexError as err:
      err_msg = "Missing script author"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def _get_version(self):
    try:
      version = [line.split(':')[1].strip() for line in ScriptInfo.get_doc().splitlines() if line.strip().startswith("@version")]
      return version[0]
    except IndexError as err:
      err_msg = "Missing script version"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def _get_date(self):
    try:
      date = [line.split(':')[1].strip() for line in ScriptInfo.get_doc().splitlines() if line.strip().startswith("@date")]
      return date[0]
    except IndexError as err:
      err_msg = "Missing script date"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def _get_args(self):
    s_args = [line.split(':')[1].strip() for line in ScriptInfo.get_doc().splitlines() if line.strip().startswith("@args")]
    try:
      args = []
      for arg in s_args:
        arg = arg.split('|')
        n_arg = {'name': arg[0],
                 'type': arg[1],
                 'help': arg[2],
                 'nargs': arg[3],
                 'default': arg[4],
                 'conflicts': arg[5],
                 'implies': arg[6],
                 'required': arg[7]
                 }
        try:
          n_arg['secret'] = arg[8]
        except IndexError:
          n_arg['secret'] = 'false'
        args.append(n_arg)
      return args
    except Exception as err:
      err_msg = f"Arguments declaration error: [{err.__class__.__name__}]: {err}"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def _add_default_args(self):
    self.parser.description = f"Author: {self._get_author()}\nDate: {self._get_date()}\nVersion: {self._get_version()}\nDescription: {self._get_description()}"
    self.parser.add_argument('--version', help='Show version and exit', action='version', version=f'%(prog)s {self._get_version()}')
    self.parser.add_argument('--source-code', help='Show source code', action=SourceAction)
    self.parser.add_argument('--hist', help='Show execution history', action=HistoryAction)
    self.parser.add_argument('--log', help='Show log content', action=LogAction)
    self.parser.add_argument('--debug', help='Force log level to DEBUG', action='store_true')
    self.parser.add_argument('--no-stdout', help='Disable log output to standard output (error et critical are kept)', action='store_true')
    self.parser.add_argument('--no-log-file', help='disable log file and history', action='store_true')

  def _add_args(self):
    types = {'str': (self._store, str),
             'int': (self._store, int),
             'float': (self._store, float),
             'choice': (self._store_choice, list),
             'bool': (self._store_bool, bool)
             }
    args = self._get_args()
    for arg in args:
      try:
        types[arg['type']][0](arg, types[arg['type']][1])
      except Exception as err:
        self.logger.critical(f"Arg: {arg}")
        err_msg = f"Arguments declaration error: [{err.__class__.__name__}]: {err}"
        self.logger.critical(err_msg)
        raise ScrippyCoreError(err_msg) from err

  def _store(self, arg, arg_type=str):
    if len(arg['nargs']) > 0:
      try:
        arg['nargs'] = int(arg['nargs'])
      except ValueError as e:
        if arg['nargs'] not in ['+', '?']:
          raise e

    options = {'help': arg['help'],
               'action': 'store',
               'type': arg_type,
               'required': self.required[arg['required'].lower()]}

    if arg['default']:
      options['default'] = arg['default']
    if arg['nargs'] != 1:
      options['nargs'] = arg['nargs']
      if arg['default']:
        options['default'] = arg['default'].split(',')

    self.parser.add_argument(f"--{arg['name']}", **options)

  def _store_bool(self, arg, arg_type=bool):
    store_type = {'false': 'store_true', 'true': 'store_false'}
    if not len(arg['default']) > 0:
      arg['default'] = 'false'
    self.parser.add_argument(f"--{arg['name']}",
                             help=arg['help'],
                             action=store_type[arg['default'].lower()],
                             required=self.required[arg['required'].lower()])

  def _store_choice(self, arg, arg_type=list):
    self.parser.add_argument(f"--{arg['name']}",
                             choices=arg['default'].split(','),
                             help=arg['help'],
                             required=self.required[arg['required'].lower()])

  def _check_implies(self, args):
    declared_arg_imps = filter(lambda arg: arg['implies'], self._get_args())
    for arg in declared_arg_imps:
      arg_name = arg['name']
      for imp in arg['implies'].split(','):
        dict_args = vars(args)
        if dict_args[arg_name] and not dict_args[imp]:
          self.parser.error(f"Params --{arg_name} implies --{imp}. Please add --{imp}")

  def _check_conflicts(self, args):
    declared_arg_conflicts = filter(lambda arg: arg['conflicts'], self._get_args())
    for arg in declared_arg_conflicts:
      arg_name = arg['name']
      for conflict_arg in arg['conflicts'].split(','):
        dict_args = vars(args)
        if dict_args[arg_name] and dict_args[conflict_arg]:
          self.parser.error(f"Params --{arg_name} conflicts with --{conflict_arg}. Please remove --{arg_name} or --{conflict_arg}")

  def register_secrets(self):
    self.logger.info("[+] Secrets registration (args)")
    declared_arg_secrets = filter(lambda arg: arg['secret'], self._get_args())
    for arg in declared_arg_secrets:
      if arg["secret"].lower() == "true":
        self.logger.add_secret(secret=getattr(self.args, arg["name"]))




import re


class Vault:
  """The Vault class manages sensitive information that should not appear in logs."""

  def __init__(self):
    self.secrets = []
    self.pattern = None

  def add(self, secret):
    if not self.is_secret(value=str(secret)):
      self.secrets.append(str(secret))
    self.pattern = re.compile(f"{'|'.join(self.secrets)}")

  def is_secret(self, value):
    return str(value) in self.secrets

  def protect(self, msg):
    if self.pattern is not None:
      return re.sub(self.pattern, "*******", str(msg))
    return msg



"""
This module provides an interface for configuring scripts based on the Scrippy framework.

This interface allows for:
- Loading and automatic consideration of a configuration file
- Verification of configuration validity
"""

import os
import configparser
from scrippy_core.scriptinfo import ScriptInfo
from scrippy_core.error_handler import ScrippyCoreError
from scrippy_core.context import GlobalContext


class Config:
  """
  Config is the main class of the scrippy_core.conf module.

  It allows for the automatic loading of the configuration and, if the script's docstring includes a set of lines starting with @conf, it enables the validation of the configuration from the configuration file.

  The lines in the configuration format declarations must adhere to the following format:

  @conf:<section>|<key>|<value_type>

  <value_type> must be one of the following recognized types:

      str (string)
      int (integer)
      float (floating-point number)
      bool (boolean)

  Example:
  --------
  Starting from the following declaration:

  @conf:log/level/str
  @conf:database/port/int
  @conf:sql/verbose/bool

  The following configuration will be checked:
  [log]
    level = ERROR
  [database]
    port = 5432
  [sql]
    verbose = True
  """

  def __init__(self):
    """Initialise l'instance.

    No validation of parameter values is performed.
    Only the configuration structure (presence of sections and keys described by the @conf set) is checked. Validation of the declared key types is also performed.
    """
    env = GlobalContext().get_env()
    self.logger = GlobalContext().get_logger()
    self.conf = configparser.ConfigParser()
    self.context = GlobalContext().get_context()
    self.doc = ScriptInfo().get_doc()
    conf_fname = f"{ScriptInfo().get_name()}.conf"
    self.config_filename = os.path.join(env.get("confdir"), conf_fname)
    if os.path.isfile(self.config_filename):
      self.logger.info(f"[+] Loading configuration file {self.config_filename}")
      self.conf.read(self.config_filename)
    else:
      self.logger.info(f"[+] No configuration file {self.config_filename}")
    self._check()

  def _check(self):
    """
    Configuration check.

    This method is automatically called by the constructor if the optional 'params' argument is provided. This method should not be called other than by the constructor of the Config class.
    """
    self.logger.info("[+] Configuration check")
    conf_lines = []
    conf_lines = [line.split(':')[1].strip() for line in self.doc.splitlines() if line.strip().startswith("@conf")]
    for line in conf_lines:
      try:
        section, key, param_type, secret = line.split("|")
        value = self.get(section, key, param_type)
        if secret.lower() == "true":
          self.logger.add_secret(secret=value)
      except ValueError:
        section, key, param_type = line.split("|")
        value = self.get(section, key, param_type)
      self.logger.debug(f"[{section}][{key}] = {value} [OK]")
    return True

  def register_secrets(self):
    self.logger.info("[+] Secrets registration (conf)")
    conf_lines = []
    conf_lines = [line.split(':')[1].strip() for line in self.doc.splitlines() if line.strip().startswith("@conf")]
    for line in conf_lines:
      try:
        section, key, param_type, secret = line.split("|")
        value = self.get(section, key, param_type)
        if secret.lower() == "true":
          self.logger.add_secret(secret=value)
      except ValueError:
        pass

  def get_section(self, section):
    """
    Returns the set of key/value pairs for the section passed as an argument.
    Raises an exception if the requested section does not exist.
    """
    sec = {}
    try:
      for key in self.conf[section]:
        sec[key] = self.conf[section][key]
      return sec
    except configparser.NoSectionError as err:
      err_msg = f"Missing or unknown configuration section: [{section}]"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def get(self, section, key, param_type='str'):
    """
    Returns the value of a key from the key name and section passed as arguments.

    Unless the 'param_type' parameter is set to one of the allowed values (str by default, int, float, or bool), the returned type is always a string.
    """
    try:
      return {'int': self.conf.getint,
              'float': self.conf.getfloat,
              'bool': self.conf.getboolean,
              'str': self.conf.get}[param_type](section, key)
    except configparser.NoOptionError as err:
      err_msg = f"Missing or unknown configuration parameter: [{section}][{key}]"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    except configparser.NoSectionError as err:
      err_msg = f"Missing or unknown configuration section: [{section}]"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    except ValueError as err:
      err_msg = f"Type error: [{section}][{key}] is not of type '{param_type}'"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    except KeyError as err:
      err_msg = f"Usage error: Unknown parameter type: '{param_type}'"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def has(self, section, key):
    """Returns True if the key section.key exists in the configuration."""
    return self.conf.has_option(section, key)



"""Execution history management."""
import sys
from argparse import Action, SUPPRESS
from scrippy_core.history import History
from scrippy_core.context import GlobalContext


class HistoryAction(Action):
  """Management class for execution history."""
  def __init__(self, option_strings,
               dest=SUPPRESS, default=SUPPRESS,
               help="Show execution history"):
    super(HistoryAction, self).__init__(option_strings=option_strings,
                                        dest=dest,
                                        default=default,
                                        nargs='?',
                                        type=int,
                                        # Default max execution saved
                                        const=10,
                                        metavar=('NB_EXECUTION (default:10)'),
                                        help=help)

  def __call__(self, parser, namespace, nb_execution, option_string=None):
    print(f"{nb_execution} last executions of {GlobalContext().get_filename()}")
    formatter = parser._get_formatter()
    formatter.add_text(History().read_history(nb_execution))
    parser._print_message(formatter.format_help(), sys.stdout)
    parser.exit()



"""Error management."""
import sys
import logging
import traceback
from scrippy_core.context import GlobalContext


class ScrippyCoreError(Exception):
  """Specific class of error."""

  def __init__(self, message):
    self.message = message
    super().__init__(self.message)


def handle_error(kind, value, tb):
  """Trap all errors and quit gracefully after all errors are logged."""
  if kind is not None:
    logger = GlobalContext().get_logger()
    logger.critical(f"[{kind.__name__}]: {str(value)}")
    if logger.getEffectiveLevel() <= logging.DEBUG:
      traceback.print_tb(tb)
    if kind != SystemExit:
      sys.exit(1)



import os
import errno
from filelock import FileLock


class PIDStack:
  """The PIDStack class allows for the management of concurrent executions."""

  def __init__(self, name, tmpdir):
    self.stack_file_path = os.path.join(tmpdir, f"{name}.pids")
    self.lock = FileLock(f"{self.stack_file_path}.lock")

  def _read(self):
    pids = []
    if os.path.isfile(self.stack_file_path):
      with open(self.stack_file_path, mode="r", encoding="utf-8") as stack_file:
        pids = [pid.strip() for pid in stack_file.readlines()]
      if pids is None:
        pids = []
    return pids

  def _write(self, pids):
    with open(self.stack_file_path, mode="w", encoding="utf-8") as stack_file:
      for pid in pids:
        stack_file.write(f"{pid}\n")

  def get_pids(self):
    """Returns the list of PIDs in queue or currently executing."""
    pids = []
    with self.lock.acquire():
      for pid in self._read():
        try:
          os.kill(int(pid), 0)
          pids.append(int(pid))
        except OSError as e:
          if e.errno == errno.EPERM:
            pids.append(pid)
          elif e.errno == errno.ESRCH:
            pass
      self._write(pids)
      return pids

  def register(self):
    """Registers a new PID in the stack"""
    with self.lock.acquire():
      pids = self.get_pids()
      pids.append(os.getpid())
      self._write(pids)

  def checkout(self):
    """Cleans the stack."""
    with self.lock.acquire():
      self._write([pid for pid in self.get_pids() if pid != str(os.getpid())])



"""Display script source."""
from rich.console import Console
from rich.syntax import Syntax
from argparse import Action, SUPPRESS
from scrippy_core.scriptinfo import ScriptInfo


class SourceAction(Action):
  """Class for code source display."""
  def __init__(self, option_strings, dest=SUPPRESS, default=SUPPRESS, help="Show source code"):
    super(SourceAction, self).__init__(option_strings=option_strings,
                                       dest=dest,
                                       default=default,
                                       nargs=0,
                                       type=bool,
                                       help=help)

  def __call__(self, parser, namespace, nb_execution, option_string=None):
    console = Console()
    syntax = Syntax.from_path(ScriptInfo.get_full_filename(), line_numbers=True)
    console.print(syntax)
    parser.exit()



import os
import time


class GlobalContext:
  _instance = None

  def __new__(cls):
    if cls._instance is None:
      cls._instance = super(GlobalContext, cls).__new__(cls)
      cls._instance._set_session()
      cls._instance.set_env(None)
      cls._instance.set_context(None)
      cls._instance.set_full_filename(None)
      cls._instance.set_filename(None)
      cls._instance.set_name(None)
      cls._instance.set_log_manager(None)
      cls._instance.set_logger(None)
    return cls._instance

  def _set_session(self):
    self.session = f"{time.time()}_{os.getpid()}"

  def get_session(self):
    return self.session

  def set_env(self, env):
    self.env = env

  def get_env(self):
    return self.env

  def set_context(self, context):
    self.context = context

  def get_context(self):
    return self.context

  def set_full_filename(self, filename):
    self.full_filename = filename

  def get_full_filename(self):
    return self.full_filename

  def set_filename(self, filename):
    self.filename = filename

  def get_filename(self):
    return self.filename

  def set_name(self, name):
    self.name = name

  def get_name(self):
    return self.name

  def set_log_manager(self, log_manager):
    self.log_manager = log_manager

  def get_log_manager(self):
    return self.log_manager

  def set_logger(self, logger):
    self.logger = logger

  def get_logger(self):
    return self.logger



"""The scrippy_core.scriptinfo sub-module."""
import os
import __main__


class ScriptInfo:
  """ScriptInfo object."""

  @staticmethod
  def get_full_filename():
    """Returns current script full path."""
    return __main__.__file__

  @staticmethod
  def get_filename():
    """Returns current script filename."""
    return os.path.basename(__main__.__file__)

  @staticmethod
  def get_name():
    """Returns current script name."""
    return os.path.splitext(ScriptInfo.get_filename())[0]

  @staticmethod
  def get_doc():
    """Returns current script doc string."""
    return __main__.__doc__



"""Scrippy Core module."""
import os
import grp
import pwd
import time
import yaml
import logging
from scrippy_core.conf import Config
from scrippy_core import error_handler
from scrippy_core.stack import PIDStack
from scrippy_core.history import History
from scrippy_core.arguments import Parser
from scrippy_core.workspace import Workspace
from scrippy_core.scriptinfo import ScriptInfo
from scrippy_core.logger.manager import LogManager
from scrippy_core.error_handler import ScrippyCoreError
from scrippy_core.context import GlobalContext

# ------------------------------------------------------------------------------
# INITIALIZATION
# ------------------------------------------------------------------------------
# Different levels of configuration files:
# - Vendor configuration file: /etc/scrippy/scrippy.yml
# - User configuration file: ~/.config/scrippy/scrippy.yml
# - SysAdmin configuration file: /usr/local/etc/scrippy/scrippy.yml
#
# Precedence:
# -----------
# SysAdmin values can not be overriden by user
# User values override Vendor values
#
conf_files = ["/etc/scrippy/scrippy.yml",
              os.path.expanduser("~/.config/scrippy/scrippy.yml"),
              "/usr/local/etc/scrippy/scrippy.yml"]

# Scrippy environment vars
env = {"logdir": None,
       "histdir": None,
       "reportdir": None,
       "tmpdir": None,
       "datadir": None,
       "templatedir": None,
       "confdir": None}

for conf_file in conf_files:
  if os.path.isfile(conf_file):
    with open(conf_file, mode="r", encoding="utf-8") as conf_file:
      scrippy_conf = yaml.load(conf_file, Loader=yaml.FullLoader)
      env["logdir"] = scrippy_conf.get("env").get("logdir")
      env["histdir"] = scrippy_conf.get("env").get("histdir")
      env["reportdir"] = scrippy_conf.get("env").get("reportdir")
      env["tmpdir"] = scrippy_conf.get("env").get("tmpdir")
      env["datadir"] = scrippy_conf.get("env").get("datadir")
      env["templatedir"] = scrippy_conf.get("env").get("templatedir")
      env["confdir"] = scrippy_conf.get("env").get("confdir")

for key, value in env.items():
  if value is None:
    raise ScrippyCoreError(f"Missing configuration key: {key}")

log_manager = LogManager()
logger = logging.getLogger("scrippy.main")
GlobalContext().set_env(env=env)
GlobalContext().set_logger(logger=logger)
GlobalContext().set_name(name=ScriptInfo().get_name())
GlobalContext().set_full_filename(filename=ScriptInfo().get_full_filename())
GlobalContext().set_filename(filename=ScriptInfo().get_filename())
GlobalContext().set_log_manager(log_manager=log_manager)
arg_parser = Parser()
if arg_parser.args.no_stdout:
  log_manager.set_log_level("error")
if arg_parser.args.debug:
  log_manager.set_log_level("debug")
arg_parser.register_secrets()


class ScriptContext:
  """Script execution context."""
  def __init__(self, retention=50, workspace=True):
    self.logger = logger
    self.filename = ScriptInfo.get_filename()
    self.name = ScriptInfo.get_name()
    self.retention = retention
    self.workspace_enabled = workspace
    self.args = arg_parser.args
    self.config = None
    self.history = None
    self.workspace = None
    self.workspace_path = None
    self.pidstack = None

  def __enter__(self):
    """Entry point."""
    self.config = Config()
    if self.config.has(section="log", key="level"):
      log_manager.set_log_level(level=self.config.get('log', 'level').lower())
    self.config.register_secrets()
    self.check_users()
    self.check_groups()
    self.stack()
    self.check_instances()
    if arg_parser.args.no_log_file:
      self.logger.warning("[+] Option --no-log-file is set -> This session will not be logged nor recorded into history")
    else:
      log_manager.add_file_handler(script=self.name,
                                   session=GlobalContext().get_session())
      self.history = History(retention=self.retention)
    if self.workspace_enabled:
      workspace_name = f"{self.name}_{GlobalContext().get_session()}"
      self.workspace = Workspace(name=workspace_name)
    if self.history is not None:
      self.history.__enter__()
    if self.workspace is not None:
      self.workspace_path = self.workspace.__enter__()
    return self

  def __exit__(self, kind, value, tb):
    """Exit point."""
    if self.workspace is not None:
      self.workspace.__exit__(kind, value, tb)
    if self.history is not None:
      self.history.__exit__(kind, value, tb)
    self.unstack()
    error_handler.handle_error(kind, value, tb)

  def check_users(self):
    """
    Checks if the current user is authorized to execute the script by comparing the "@user" declarations in the script's header.

    Raises a ScrippyCoreError if the user is not authorized.
    """
    doc = ScriptInfo().get_doc()
    user_id = os.geteuid()
    try:
      users = [line.split(':')[1].strip() for line in doc.splitlines() if line.strip().startswith("@user")]
      users = [pwd.getpwnam(user).pw_uid for user in users]
      if len(users) > 0 and user_id not in users:
        raise ScrippyCoreError('[BadUserError] Unauthorized user')
    except KeyError as err:
      raise ScrippyCoreError(f"[UnknownUserError] Unknown user: {str(err)}") from err
    except Exception as err:
      err_msg = f"[{err.__class__.__name__}] {err}"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def check_groups(self):
    """
    Verifies that the current user is authorized to execute the script by comparing the "@group" declarations in the script's header.

    Raises a ScrippyCoreError if the user is not part of the authorized groups.
    """
    doc = ScriptInfo().get_doc()
    user_groups = os.getgroups()
    try:
      groups = [line.split(':')[1].strip() for line in doc.splitlines() if line.strip().startswith("@group")]
      groups = [grp.getgrnam(group)[2] for group in groups]
      if len(groups) > 0:
        if not len([groups for g in user_groups if g in groups]) > 0:
          raise ScrippyCoreError("[BadGroupError] User is not member of an authorized group")
    except KeyError as err:
      raise ScrippyCoreError(f"[UnknownGroupError] Unknown group: {str(err)}") from err
    except Exception as err:
      err_msg = f"[{err.__class__.__name__}] {err}"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def check_instances(self):
    """
    Verifies that the maximum number of allowed instances to be simultaneously executed is not reached by comparing the declaration @max_instance in the script's header with the number of PIDs returned by the pidstack.

    If the maximum instance limit is reached, the script will pause until the number of currently running instances is lower than the allowed number of instances.

    When the maximum allowed instances are not reached, the script registers itself in the pidstack and proceeds with execution.
    """
    doc = ScriptInfo().get_doc()
    sleep_step = 3
    bools = ["true", "1", "on"]
    try:
      max_instance = [line.split(':')[1].strip() for line in doc.splitlines() if line.strip().startswith("@max_instance")][0]
    except IndexError:
      max_instance = 0
    try:
      timeout = [line.split(':')[1].strip() for line in doc.splitlines() if line.strip().startswith("@timeout")][0]
    except IndexError:
      timeout = 0
    try:
      exit_on_wait = [line.split(':')[1].strip() for line in doc.splitlines() if line.strip().startswith("@exit_on_wait")][0]
    except IndexError:
      exit_on_wait = "False"
    try:
      exit_on_timeout = [line.split(':')[1].strip() for line in doc.splitlines() if line.strip().startswith("@exit_on_timeout")][0]
    except IndexError:
      exit_on_timeout = "False"
    pids = self.pidstack.get_pids()
    try:
      timeout = int(timeout)
      max_instance = int(max_instance)
      exit_on_timeout = exit_on_timeout.lower() in bools
      exit_on_wait = exit_on_wait.lower() in bools
      if max_instance > 0 and len(pids) > max_instance:
        self.logger.info(f"[+] Waiting for an execution slot: {len(pids)}/{max_instance} [{timeout}s]")
        while len(pids) > max_instance and pids[0] != os.getpid():
          timeout -= sleep_step
          if timeout <= 0 and exit_on_timeout:
            raise ScrippyCoreError("TimeoutError: Timeout expired")
          if exit_on_wait:
            raise ScrippyCoreError(f"EagernessError: `exit_on_wait` is set to {exit_on_wait}")
          pids = self.pidstack.get_pids()
          time.sleep(sleep_step)
    except Exception as err:
      err_msg = f"[{err.__class__.__name__}] {err}"
      self.logger.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def stack(self):
    try:
      self.pidstack = PIDStack(name=self.name,
                               tmpdir=GlobalContext().get_env().get("tmpdir"))
      self.pidstack.register()
    except PermissionError as err:
      err_msg = f"Error while creating instances stack: [{err.__class__.__name__}] {err}"
      self.logger.critical(err_msg)
      raise err

  def unstack(self):
    self.pidstack.checkout()



"""The scrippy_core.history sub-module provides the necessary objects for managing execution history."""
import os
import sys
import time
import sqlite3
import getpass
from datetime import datetime
from prettytable import PrettyTable
from scrippy_core.context import GlobalContext


class History:
  """L'objet History est l'objet qui fournit l'ensemble des méthodes permettant historisation des executions."""
  def __init__(self, retention=90):
    self.logger = GlobalContext().get_logger()
    self.logger.info(f"[+] Opening history (retention={retention})")
    env = GlobalContext().get_env()
    self.session = GlobalContext().get_session()
    self.retention = retention
    history_db = f"{GlobalContext().get_name()}.db"
    self.hist_filename = os.path.join(env.get("histdir"), history_db)
    self.user_exec = getpass.getuser()
    self.start = time.time()
    self.user_origin = os.getenv("SUDO_USER") or getpass.getuser()
    self.params = sys.argv[1:]

  def __enter__(self):
    """Entry point."""
    if os.path.isfile(self.hist_filename):
      self.purge_old_history()
    else:
      self.create_history()
    session_exec_params = [self.session,
                           self.start,
                           self.user_origin,
                           self.user_exec,
                           str(self.params)]
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("insert into executions (session, start, origin, exec, params) values (?, ?, ?, ?, ?)", session_exec_params)
      conn.commit()

  def __exit__(self, kind, value, traceback):
    """Exit point."""
    exit_code = 0
    if kind == SystemExit:
      if value.code is not None:
        exit_code = value.code
    elif kind is not None:
      exit_code = kind.__name__
    self.close_history(exit_code)

  def create_history(self):
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("CREATE TABLE executions (session TEXT, start REAL, end REAL, duration REAL, origin TEXT, exec TEXT, code TEXT, params TEXT, exit_code INTEGER)")
      conn.commit()

  def purge_old_history(self):
    """Purges the history file."""
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("delete from executions where session not in (select session from executions order by end DESC limit ?)", [self.retention])
      conn.commit()

  def close_history(self, exit_code):
    """Record session exit code and duration."""
    end = time.time()
    delta = datetime.fromtimestamp(end) - datetime.fromtimestamp(self.start)
    result = [end,
              self._format_time_delta(delta),
              exit_code,
              self.session]
    with sqlite3.connect(self.hist_filename) as conn:
      cursor = conn.cursor()
      cursor.execute("update executions set end=?, duration=?, exit_code=? where session=?", result)
      conn.commit()
    self.logger.info(f"[+] End: {exit_code}")

  def get_last_session(self):
    if os.path.isfile(self.hist_filename):
      with sqlite3.connect(self.hist_filename) as conn:
        cursor = conn.cursor()
        cursor.execute("select session from executions order by start desc limit 1")
        return cursor.fetchone()[0]
    return None

  def read_history(self, nb_execution):
    """Read and displays history."""
    if os.path.isfile(self.hist_filename):
      table = PrettyTable()
      table.field_names = ["Session",
                           "Start",
                           "End",
                           "Duration",
                           "Origin",
                           "Exec",
                           "Code",
                           "Params",
                           "Exit"]
      with sqlite3.connect(self.hist_filename) as conn:
        cursor = conn.cursor()
        cursor.execute("select * from executions order by start desc limit ?", [nb_execution])
        sessions = []
        for session in cursor.fetchall():
          sessions.append(self._get_human_readable_dates(session))
        table.add_rows(sessions)
        return table.get_string()
    return None

  def _get_human_readable_dates(self, session):
    """Return session records with human readable start and end time"""
    session = list(session)
    session[1] = datetime.fromtimestamp(session[1]).strftime("%d/%m/%Y %H:%M:%S")
    session[2] = datetime.fromtimestamp(session[2]).strftime("%d/%m/%Y %H:%M:%S")
    return session

  def _format_time_delta(self, tdelta):
    """Take a timedelta object and formats it for humans.
    From https://gist.github.com/dhrrgn/7255361
    """
    delta = dict(days=tdelta.days)
    delta["hrs"], rem = divmod(tdelta.seconds, 3600)
    delta["min"], delta["sec"] = divmod(rem, 60)
    if delta["min"] == 0:
      fmt = "{sec} sec"
    elif delta["hrs"] == 0:
      fmt = "{min} min {sec} sec"
    elif delta["days"] == 0:
      fmt = "{hrs} hr(s) {min} min {sec} sec"
    else:
      fmt = "{days} day(s) {hrs} hr(s) {min} min {sec} sec"
    return fmt.format(**delta)



from argparse import Action, SUPPRESS
from scrippy_core.history import History
from scrippy_core.context import GlobalContext


class LogAction(Action):
  def __init__(self,
               option_strings,
               dest=SUPPRESS,
               default=SUPPRESS,
               help="Show execution history"):
    super(LogAction, self).__init__(option_strings=option_strings,
                                    dest=dest,
                                    default=default,
                                    nargs='?',
                                    type=str,
                                    metavar=('SESSION (default: last)'),
                                    help=help)

  def __call__(self, parser, namespace, session, option_string=None):
    session = session or History().get_last_session()
    GlobalContext().get_log_manager().print_logfile(GlobalContext().get_name(),
                                                    session)
    parser.exit()
